﻿/*
PaletteTileCodGen: Utility to generate palette and tile assembly code for Sega Genesis
Borrows heavily from https://github.com/huguesjohnson/MDLib
Copyright (c) 2015 Hugues Johnson

PaletteTileCodGen is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 
as published by the Free Software Foundation.

PaletteTileCodGen is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

using com.huguesjohnson.MegaDriveIO;
using System.IO;

namespace PaletteTileCodeGen
{
    public partial class MainForm:Form
    {
		private System.Windows.Forms.RadioButton selectedButton;
		private int selectedPaletteIndex;
		private Palette palette;
		private MDTile tile;
		private int selectedPixelX;
		private int selectedPixelY;
		private int TILE_PIXEL_ROWS=32;
		private int TILE_PIXEL_COLS=32;
		private int TILE_PIXEL_WIDTH;
		private int TILE_PIXEL_HEIGHT;
        private bool suspendSliderValueChange=false;

        public MainForm()
        {
            InitializeComponent();
		    this.TILE_PIXEL_WIDTH=this.pictureBoxTiles.Width/TILE_PIXEL_COLS;
		    this.TILE_PIXEL_HEIGHT=this.pictureBoxTiles.Height/TILE_PIXEL_ROWS;
            this.palette=new Palette();
            this.palette.Entries[0]=new PaletteEntry(3822);
            for(int i=1;i<16;i++)
            { 
                this.palette.Entries[i]=new PaletteEntry(0);
            }
            this.setPaletteButtonColors();
			//select the first button
			this.radioButtonPalette0.Checked=true;
			radioButtonPalette0_CheckedChanged(null,null);
            //generate the default palette code               
            this.generatePaletteCode();
            //create the default tile
            this.tile=new MDTile();
            this.tile.PixelData=new byte[this.TILE_PIXEL_COLS][];
            for(int i=0;i<this.TILE_PIXEL_COLS;i++)
            {
                this.tile.PixelData[i]=new byte[this.TILE_PIXEL_ROWS];
            }
            this.generateTileCode();
        }

        private void setPaletteButtonColors()
        { 
			this.radioButtonPalette0.BackColor=palette.Entries[0].ToRGB();
			this.radioButtonPalette0.ForeColor=this.getInverseColor(this.radioButtonPalette0.BackColor); 
			this.radioButtonPalette1.BackColor=palette.Entries[1].ToRGB();
			this.radioButtonPalette1.ForeColor=this.getInverseColor(this.radioButtonPalette1.BackColor); 
			this.radioButtonPalette2.BackColor=palette.Entries[2].ToRGB();
			this.radioButtonPalette2.ForeColor=this.getInverseColor(this.radioButtonPalette2.BackColor); 
			this.radioButtonPalette3.BackColor=palette.Entries[3].ToRGB();
			this.radioButtonPalette3.ForeColor=this.getInverseColor(this.radioButtonPalette3.BackColor); 
			this.radioButtonPalette4.BackColor=palette.Entries[4].ToRGB();
			this.radioButtonPalette4.ForeColor=this.getInverseColor(this.radioButtonPalette4.BackColor); 
			this.radioButtonPalette5.BackColor=palette.Entries[5].ToRGB();
			this.radioButtonPalette5.ForeColor=this.getInverseColor(this.radioButtonPalette5.BackColor); 
			this.radioButtonPalette6.BackColor=palette.Entries[6].ToRGB();
			this.radioButtonPalette6.ForeColor=this.getInverseColor(this.radioButtonPalette6.BackColor); 
			this.radioButtonPalette7.BackColor=palette.Entries[7].ToRGB();
			this.radioButtonPalette7.ForeColor=this.getInverseColor(this.radioButtonPalette7.BackColor); 
			this.radioButtonPalette8.BackColor=palette.Entries[8].ToRGB();
			this.radioButtonPalette8.ForeColor=this.getInverseColor(this.radioButtonPalette8.BackColor); 
			this.radioButtonPalette9.BackColor=palette.Entries[9].ToRGB();
			this.radioButtonPalette9.ForeColor=this.getInverseColor(this.radioButtonPalette9.BackColor); 
			this.radioButtonPalette10.BackColor=palette.Entries[10].ToRGB();
			this.radioButtonPalette10.ForeColor=this.getInverseColor(this.radioButtonPalette10.BackColor); 
			this.radioButtonPalette11.BackColor=palette.Entries[11].ToRGB();
			this.radioButtonPalette11.ForeColor=this.getInverseColor(this.radioButtonPalette11.BackColor); 
			this.radioButtonPalette12.BackColor=palette.Entries[12].ToRGB();
			this.radioButtonPalette12.ForeColor=this.getInverseColor(this.radioButtonPalette12.BackColor); 
			this.radioButtonPalette13.BackColor=palette.Entries[13].ToRGB();
			this.radioButtonPalette13.ForeColor=this.getInverseColor(this.radioButtonPalette13.BackColor); 
			this.radioButtonPalette14.BackColor=palette.Entries[14].ToRGB();
			this.radioButtonPalette14.ForeColor=this.getInverseColor(this.radioButtonPalette14.BackColor); 
			this.radioButtonPalette15.BackColor=palette.Entries[15].ToRGB();
    		this.radioButtonPalette15.ForeColor=this.getInverseColor(this.radioButtonPalette15.BackColor); 
        }

		private Color[] getPalette()
		{
			Color[] palette=new Color[16];
			palette[0]=this.radioButtonPalette0.BackColor;
			palette[1]=this.radioButtonPalette1.BackColor;
			palette[2]=this.radioButtonPalette2.BackColor;
			palette[3]=this.radioButtonPalette3.BackColor;
			palette[4]=this.radioButtonPalette4.BackColor;
			palette[5]=this.radioButtonPalette5.BackColor;
			palette[6]=this.radioButtonPalette6.BackColor;
			palette[7]=this.radioButtonPalette7.BackColor;
			palette[8]=this.radioButtonPalette8.BackColor;
			palette[9]=this.radioButtonPalette9.BackColor;
			palette[10]=this.radioButtonPalette10.BackColor;
			palette[11]=this.radioButtonPalette11.BackColor;
			palette[12]=this.radioButtonPalette12.BackColor;
			palette[13]=this.radioButtonPalette13.BackColor;
			palette[14]=this.radioButtonPalette14.BackColor;
			palette[15]=this.radioButtonPalette15.BackColor;
			return(palette);
		}

		//this doesn't really do what the name implies
		private Color getInverseColor(Color c)
		{
			if((c.R>200)&&(c.G>200)&&(c.B>200))
			{
				return(Color.Black);
			}
			else
			{
				return(Color.White);
			}
		}

		private void onButtonClick(int newPaletteIndex)
		{
            this.suspendSliderValueChange=true;
			this.selectedPaletteIndex=newPaletteIndex;
			this.trackBarR.Value=this.palette.Entries[this.selectedPaletteIndex].R;
			this.trackBarG.Value=this.palette.Entries[this.selectedPaletteIndex].G;
			this.trackBarB.Value=this.palette.Entries[this.selectedPaletteIndex].B;
            this.suspendSliderValueChange=false;
		}

        private void radioButtonPalette0_CheckedChanged(object sender,EventArgs e)
        {
            if(!this.radioButtonPalette0.Checked){return;}
            this.selectedButton=this.radioButtonPalette0;
			this.onButtonClick(0);
        }

		private void radioButtonPalette1_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette1.Checked){return;}
			this.selectedButton=this.radioButtonPalette1;
			this.onButtonClick(1);
		}

		private void radioButtonPalette2_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette2.Checked){return;}
			this.selectedButton=this.radioButtonPalette2;
			this.onButtonClick(2);
		}

		private void radioButtonPalette3_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette3.Checked){return;}
			this.selectedButton=this.radioButtonPalette3;
			this.onButtonClick(3);
		}

		private void radioButtonPalette4_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette4.Checked){return;}
			this.selectedButton=this.radioButtonPalette4;
			this.onButtonClick(4);
		}

		private void radioButtonPalette5_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette5.Checked){return;}
			this.selectedButton=this.radioButtonPalette5;
			this.onButtonClick(5);
		}

		private void radioButtonPalette6_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette6.Checked){return;}
			this.selectedButton=this.radioButtonPalette6;
			this.onButtonClick(6);
		}

		private void radioButtonPalette7_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette7.Checked){return;}
			this.selectedButton=this.radioButtonPalette7;
			this.onButtonClick(7);
		}

		private void radioButtonPalette8_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette8.Checked){return;}
			this.selectedButton=this.radioButtonPalette8;
			this.onButtonClick(8);
		}

		private void radioButtonPalette9_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette9.Checked){return;}
			this.selectedButton=this.radioButtonPalette9;
			this.onButtonClick(9);
		}

		private void radioButtonPalette10_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette10.Checked){return;}
			this.selectedButton=this.radioButtonPalette10;
			this.onButtonClick(10);
		}

		private void radioButtonPalette11_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette11.Checked){return;}
			this.selectedButton=this.radioButtonPalette11;
			this.onButtonClick(11);
		}

		private void radioButtonPalette12_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette12.Checked){return;}
			this.selectedButton=this.radioButtonPalette12;
			this.onButtonClick(12);
		}

		private void radioButtonPalette13_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette13.Checked){return;}
			this.selectedButton=this.radioButtonPalette13;
			this.onButtonClick(13);
		}

		private void radioButtonPalette14_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette14.Checked){return;}
			this.selectedButton=this.radioButtonPalette14;
			this.onButtonClick(14);
		}

		private void radioButtonPalette15_CheckedChanged(object sender, System.EventArgs e)
		{
            if(!this.radioButtonPalette15.Checked){return;}
			this.selectedButton=this.radioButtonPalette15;
			this.onButtonClick(15);
		}

        private void trackBar_ValueChanged(object sender,System.EventArgs e)
		{
            if(suspendSliderValueChange){return;}
			this.palette.Entries[this.selectedPaletteIndex].R=(ushort)this.trackBarR.Value;
			this.palette.Entries[this.selectedPaletteIndex].G=(ushort)this.trackBarG.Value;
			this.palette.Entries[this.selectedPaletteIndex].B=(ushort)this.trackBarB.Value;
			this.selectedButton.BackColor=palette.Entries[this.selectedPaletteIndex].ToRGB();
			this.selectedButton.ForeColor=this.getInverseColor(this.selectedButton.BackColor); 
            this.generatePaletteCode();
            this.pictureBoxTiles.Refresh();
		}

        //this is not super-efficient - if performance is ever an issue refactor so only changed entries are updated
        private void generatePaletteCode()
        { 
            StringBuilder builder=new StringBuilder();
            for(int i=0;i<16;i++) 
            { 
                PaletteEntry entry=this.palette.Entries[i];
				builder.Append("\tdc.w\t$");
                builder.Append(entry.ToHexString());
                builder.Append("\t; ");
                Color c=entry.ToRGB();
                builder.Append("rgb(");
                builder.Append(c.R.ToString());
                builder.Append(",");
                builder.Append(c.G.ToString());
                builder.Append(",");
                builder.Append(c.B.ToString());
                builder.Append(") => ");
                builder.Append(entry.ToBinaryString());
                builder.Append(" => ");
                builder.Append(entry.ToHexString());
                builder.Append(Environment.NewLine);
            }
            this.textBoxPaletteCode.Text=builder.ToString();
        }

        //this is also not super-efficient - if performance is ever an issue refactor so only changed entries are updated
        private void generateTileCode()
        { 
            byte[][] pixelData=this.tile.PixelData;
            StringBuilder builder=new StringBuilder();
            for(int row=0;row<this.TILE_PIXEL_ROWS;row++) 
            {
                builder.Append("\tdc.l\t$");
                for(int col=0;col<8;col++) 
                {
                    if(pixelData[col][row]<10){builder.Append(pixelData[col][row]);}
                    else if(pixelData[col][row]==10){builder.Append('A');} 
                    else if(pixelData[col][row]==11){builder.Append('B');} 
                    else if(pixelData[col][row]==12){builder.Append('C');} 
                    else if(pixelData[col][row]==13){builder.Append('D');} 
                    else if(pixelData[col][row]==14){builder.Append('E');} 
                    else if(pixelData[col][row]==15){builder.Append('F');} 
                }            
                builder.Append(Environment.NewLine);
                if(((row+1)%8)==0){builder.Append(Environment.NewLine);}
            }
            for(int row=0;row<this.TILE_PIXEL_ROWS;row++) 
            {
                builder.Append("\tdc.l\t$");
                for(int col=8;col<16;col++) 
                {
                    if(pixelData[col][row]<10){builder.Append(pixelData[col][row]);}
                    else if(pixelData[col][row]==10){builder.Append('A');} 
                    else if(pixelData[col][row]==11){builder.Append('B');} 
                    else if(pixelData[col][row]==12){builder.Append('C');} 
                    else if(pixelData[col][row]==13){builder.Append('D');} 
                    else if(pixelData[col][row]==14){builder.Append('E');} 
                    else if(pixelData[col][row]==15){builder.Append('F');} 
                }            
                builder.Append(Environment.NewLine);
                if(((row+1)%8)==0){builder.Append(Environment.NewLine);}
            }
            for(int row=0;row<this.TILE_PIXEL_ROWS;row++) 
            {
                builder.Append("\tdc.l\t$");
                for(int col=16;col<24;col++) 
                {
                    if(pixelData[col][row]<10){builder.Append(pixelData[col][row]);}
                    else if(pixelData[col][row]==10){builder.Append('A');} 
                    else if(pixelData[col][row]==11){builder.Append('B');} 
                    else if(pixelData[col][row]==12){builder.Append('C');} 
                    else if(pixelData[col][row]==13){builder.Append('D');} 
                    else if(pixelData[col][row]==14){builder.Append('E');} 
                    else if(pixelData[col][row]==15){builder.Append('F');} 
                }            
                builder.Append(Environment.NewLine);
                if(((row+1)%8)==0){builder.Append(Environment.NewLine);}
            }
            for(int row=0;row<this.TILE_PIXEL_ROWS;row++) 
            {
                builder.Append("\tdc.l\t$");
                for(int col=24;col<32;col++) 
                {
                    if(pixelData[col][row]<10){builder.Append(pixelData[col][row]);}
                    else if(pixelData[col][row]==10){builder.Append('A');} 
                    else if(pixelData[col][row]==11){builder.Append('B');} 
                    else if(pixelData[col][row]==12){builder.Append('C');} 
                    else if(pixelData[col][row]==13){builder.Append('D');} 
                    else if(pixelData[col][row]==14){builder.Append('E');} 
                    else if(pixelData[col][row]==15){builder.Append('F');} 
                }            
                builder.Append(Environment.NewLine);
                if(((row+1)%8)==0){builder.Append(Environment.NewLine);}
            }
            this.textBoxTileCode.Text=builder.ToString();
        }

        private void buttonCopyPaletteToClipboard_Click(object sender,EventArgs e)
        {
            Clipboard.Clear();
            Clipboard.SetText(this.textBoxPaletteCode.Text);
        }

        private void pictureBoxTiles_Paint(object sender,PaintEventArgs e)
        {
            Graphics g=e.Graphics;
			Color[] palette=this.getPalette();
			int xCounter=0;
			int yCounter=0;
			for(int y=0;y<this.TILE_PIXEL_COLS;y++)
			{
				for(int x=0;x<this.TILE_PIXEL_ROWS;x++)
				{
					int pixel=this.tile.PixelData[x][y];
					int xc=(xCounter*TILE_PIXEL_WIDTH);
					int yc=(yCounter*TILE_PIXEL_HEIGHT);
					SolidBrush brush=new SolidBrush(palette[pixel]); 
					e.Graphics.FillRectangle(brush,xc,yc,TILE_PIXEL_WIDTH,TILE_PIXEL_HEIGHT);
					//increment counters
					if(xCounter==(this.TILE_PIXEL_ROWS-1))
					{ 
						xCounter=0; 
						yCounter++;
					}
					else
					{
						xCounter++;
					}
				}
			}
            //draw gridlines
			Pen pen=new Pen(Color.LightGray);
			int width=this.pictureBoxTiles.Width;
			int height=this.pictureBoxTiles.Height;
            int lineCounter=1;
			for(int index=TILE_PIXEL_WIDTH;index<width;index+=TILE_PIXEL_WIDTH)
			{
                if((lineCounter%8)==0){pen.Color=Color.Black;}
				g.DrawLine(pen,index,0,index,height);
                if((lineCounter%8)==0){pen.Color=Color.LightGray;}
                lineCounter++;
			}
            lineCounter=1;
			for(int index=TILE_PIXEL_HEIGHT;index<height;index+=TILE_PIXEL_HEIGHT)
			{
                if((lineCounter%8)==0){pen.Color=Color.Black;}
				g.DrawLine(pen,0,index,width,index);
                if((lineCounter%8)==0){pen.Color=Color.LightGray;}
                lineCounter++;
			}
			base.OnPaint(e);
        }

        private void pictureBoxTiles_MouseMove(object sender,MouseEventArgs e)
        {
			int pixelColumn=e.X/TILE_PIXEL_WIDTH;
			int pixelRow=e.Y/TILE_PIXEL_HEIGHT;
			int selectedPixel=pixelColumn+(this.TILE_PIXEL_COLS*pixelRow);
			if((selectedPixel>=0)&&(selectedPixel<(this.TILE_PIXEL_COLS*this.TILE_PIXEL_ROWS))) //this check *should* never fail
			{
				if((pixelColumn<0)||(pixelColumn>(this.TILE_PIXEL_COLS-1))||(pixelRow<0)||(pixelRow>(this.TILE_PIXEL_ROWS-1))){return;}
				this.selectedPixelX=pixelColumn;
				this.selectedPixelY=pixelRow;
				if(e.Button.Equals(MouseButtons.Left))
				{
					this.tile.PixelData[this.selectedPixelX][this.selectedPixelY]=(byte)(this.selectedPaletteIndex);
					this.pictureBoxTiles.Refresh();
                    this.generateTileCode();
				}			
			}
        }

        private void pictureBoxTiles_MouseUp(object sender,MouseEventArgs e)
        {
    		//figure out which tile was clicked
			int pixelColumn=e.X/TILE_PIXEL_WIDTH;
			int pixelRow=e.Y/TILE_PIXEL_HEIGHT;
			int selectedPixel=pixelColumn+((this.TILE_PIXEL_COLS*pixelRow)*pixelRow);
			if((selectedPixel>=0)&&(selectedPixel<(this.TILE_PIXEL_COLS*this.TILE_PIXEL_ROWS))) //this check *should* never fail
			{
				if((pixelColumn<0)||(pixelColumn>(this.TILE_PIXEL_COLS-1))||(pixelRow<0)||(pixelRow>(this.TILE_PIXEL_ROWS-1))){ return; }
				this.selectedPixelX=pixelColumn;
				this.selectedPixelY=pixelRow;
				if(e.Button.Equals(MouseButtons.Left))
				{
					this.tile.PixelData[this.selectedPixelX][this.selectedPixelY]=(byte)(this.selectedPaletteIndex);
					this.pictureBoxTiles.Refresh();
                    this.generateTileCode();
				}
			}
        }

        private void buttonCopyTileCodeToClipboard_Click(object sender,EventArgs e)
        {
            Clipboard.Clear();
            Clipboard.SetText(this.textBoxTileCode.Text);
        }

        private void buttonSavePalette_Click(object sender,EventArgs e)
        {
            this.saveFile(this.textBoxPaletteCode.Text);
        }

        private void buttonSaveTiles_Click(object sender,EventArgs e)
        {
            this.saveFile(this.textBoxTileCode.Text);
        }

        private void saveFile(string textToSave)
        {
            if(this.saveFileDialog.ShowDialog()==DialogResult.OK)
            {
                using(StreamWriter writer=new StreamWriter(this.saveFileDialog.FileName))
                writer.Write(textToSave);
            }
        }

        private void buttonOpenPalette_Click(object sender,EventArgs e)
        {
            if(this.openFileDialog.ShowDialog()==DialogResult.OK)
            {
                try
                {
                    StreamReader reader=new StreamReader(this.openFileDialog.FileName);
                    string line;
                    int paletteIndex=0;
                    while((line=reader.ReadLine())!=null)
                    {
                        //I suppose a regular expression would be better
                        int indexOf=line.IndexOf('$');
                        if(indexOf>0)
                        {
                            string hexString=line.Substring(indexOf+1,4);
                            ushort value=(ushort)Convert.ToInt32(hexString,16);
                            this.palette.Entries[paletteIndex]=new PaletteEntry(value);
                            paletteIndex++;
                        }
                    }
                    this.setPaletteButtonColors();
                    this.generatePaletteCode();
                    this.pictureBoxTiles.Refresh();                  
                }
                catch(Exception x)
                {
                    MessageBox.Show(x.Message);
                }
            }
        }

        private void buttonOpenTiles_Click(object sender,EventArgs e)
        {
            if(this.openFileDialog.ShowDialog()==DialogResult.OK)
            { 
                try
                {
                    StreamReader reader=new StreamReader(this.openFileDialog.FileName);
                    string line;
                    int tileColBase=0;
                    int tileRow=0;
                    while((line=reader.ReadLine())!=null)
                    {
                        //I suppose a regular expression would be better
                        int indexOf=line.IndexOf('$');
                        if(indexOf>0)
                        {
                            string tileString=line.Substring(indexOf+1,8);
                            for(int index=0;index<8;index++) 
                            { 
                                string digit=tileString.Substring(index,1);
                                byte value=(byte)Convert.ToInt32(digit,16);
                                this.tile.PixelData[tileColBase+index][tileRow]=value;
                            }
                            tileRow++;
                            if(tileRow==this.TILE_PIXEL_ROWS)
                            {
                                tileRow=0;
                                tileColBase+=8;
                            }
                        }
                    }
                    this.pictureBoxTiles.Refresh();                  
                    this.generateTileCode();
                }
                catch(Exception x)
                {
                    MessageBox.Show(x.Message);
                }
            }
        }
    }
}
